Package: BiocHail
Date: 2025-06-30
Title: basilisk and hail
Version: 1.12.0
Authors@R: 
    c(person(
        "Vincent", "Carey", role = c("aut", "cre"),
        email = "stvjc@channing.harvard.edu",
        comment = c(ORCID = "0000-0003-4046-0063")
     )
    )
Description: Use hail via basilisk when appropriate, or via reticulate.
	This package can be used in terra.bio to interact with UK Biobank
	resources processed by hail.is.
License: Artistic-2.0
Encoding: UTF-8
Depends: R (>= 4.3.0), graphics, stats, utils
Imports: reticulate, basilisk, BiocFileCache, methods, dplyr,
        BiocGenerics
Suggests: knitr, testthat, BiocStyle, ggplot2, DT
VignetteBuilder: knitr
biocViews: Infrastructure
RoxygenNote: 7.3.2
URL: https://github.com/vjcitn/BiocHail
BugReports: https://github.com/vjcitn/BiocHail/issues
git_url: https://git.bioconductor.org/packages/BiocHail
git_branch: RELEASE_3_23
git_last_commit: fcd2afe
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-29
NeedsCompilation: no
Packaged: 2026-04-29 20:39:32 UTC; biocbuild
Author: Vincent Carey [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-4046-0063>)
Maintainer: Vincent Carey <stvjc@channing.harvard.edu>

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,hail.table.Table)
S3method(filter,hail.table.Table)
S3method(get_key,hail.table.Table)
export(as.data.frame)
export(bare_hail)
export(filter)
export(get_1kg)
export(get_key)
export(get_ukbb_sumstat_10kloci_mt)
export(hail_init)
export(hail_init_simple)
export(hail_stop)
export(multipop_df)
export(osn_1kg_path)
export(osn_ukbb_sumst10k_path)
export(path_1kg_annotations)
export(rg_update)
export(ukbb_init)
exportMethods(colnames)
exportMethods(rownames)
import(BiocFileCache)
import(basilisk)
import(methods)
import(utils)
importFrom(BiocGenerics,as.data.frame)
importFrom(BiocGenerics,colnames)
importFrom(BiocGenerics,rownames)
importFrom(dplyr,filter)
importFrom(reticulate,import)

# BiocHail

This is a multi-step overview of Hail for genetic association
studies

## Tutorial overview

The basic layout, following the [Hail GWAS tutorial](https://hail.is/docs/0.2/tutorials/01-genome-wide-association-study.html):

![topics](https://github.com/vjcitn/BiocHail/raw/main/toc1.png)

We'll see how to use a very small excerpt from the
1000 genomes study to produce

![litman](https://github.com/vjcitn/BiocHail/raw/main/litman.png)

Along the way, we illustrate and adjust for population stratification:

![popstrat](https://github.com/vjcitn/BiocHail/raw/main/popstrat.png)

## Larger data problem -- 1000 genomes data with T2T reference

We have arranged a serialization of genotypes on chromosome 17
for 3202 1000 genomes samples.

## Relevant LD resources ... in which cloud?

Still to come.

## Simulating multipopulation genotype surveys

Still to come.

## Binding phenotypes from FHIR to Hail

Still to come.

The 1kg_annotations.txt was retrieved from https://storage.googleapis.com/hail-1kg/tutorial_data.tar

The my2ss.zip is a MatrixTable instance that has been compressed.  It is derived from the
annotation for GWAS summary statistics from UK Biobank provided at the Broad UK Biobank
portal.  Example of the underlying python structs:

[[7271]]
Struct(trait_type='prescriptions', phenocode='zopiclone', pheno_sex='both_sexes', coding='', modifier='', pheno_data=[Struct(n_cases=170, n_controls=6466, heritability=0.10794648895016906, saige_version='SAIGE_0.36.3', inv_normalized=False, pop='AFR'), Struct(n_cases=382, n_controls=8494, heritability=0.03487760637490611, saige_version='SAIGE_0.36.3', inv_normalized=False, pop='CSA'), Struct(n_cases=66, n_controls=2643, heritability=0.12246893045325642, saige_version='SAIGE_0.36.3', inv_normalized=False, pop='EAS'), Struct(n_cases=14323, n_controls=406208, heritability=0.051400442321158576, saige_version='SAIGE_0.36.3', inv_normalized=False, pop='EUR'), Struct(n_cases=65, n_controls=1534, heritability=0.06825527421937107, saige_version='SAIGE_0.36.3', inv_normalized=False, pop='MID')], description=None, description_more=None, coding_description=None, category='GABA agonist,sedative', n_cases_full_cohort_both_sexes=17242, n_cases_full_cohort_females=9864, n_cases_full_cohort_males=5441)

#' update the reference genome for a hail instance
#' @param hc hail context
#' @param init character(1) valid name for a reference genome, defaults to "GRCh38"
#' @param newjson character(1) path to a json spec of a reference genome [needs doc]
#' @return a python list; the function is used for its side effect
#' @examples
#' hl <- hail_init()
#' rg_update(hl)
#' @export
rg_update <- function(hc, init = "GRCh38", newjson = system.file("json/t2tAnVIL.json", package = "BiocHail")) {
  rg <- hc$get_reference(init)
  rg$read(newjson)
}

#https://github.com/hail-is/hail/blob/main/ci/pinned-requirements.txt 62ef
from_pinned = c("hail==0.2.135", 
"ukbb_pan_ancestry==0.0.2",   # top 2 here are not in pinned-requirements
"certifi==2025.4.26",
"cffi==1.17.1",
"charset-normalizer==3.4.2",
"click==8.1.8",
"cryptography==45.0.4",
"distro==1.9.0",
"gidgethub==5.4.0",
"idna==3.10",
"pycparser==2.22",
"pyjwt==2.10.1",
"requests==2.32.4",
"typing-extensions==4.14.0",
"uritemplate==4.2.0",
"urllib3==1.26.20",
"zulip==0.9.0")

# necessary for python module control
bsklenv <- basilisk::BasiliskEnvironment(
  envname = "bsklenv", packages = "python==3.9.22",
  pkgname = "BiocHail", pip = from_pinned
)

#' Open Storage Network path to a zip of hail MatrixTable with some 1kg data for the Hail.is GWAS tutorial
#' @return character(1) URL to zip
#' @examples
#' osn_1kg_path()
#' @export
osn_1kg_path <- function() "https://bir190004-bucket01.mghp.osn.xsede.org/BiocHailData/1kg.zip"

#' interface to 1kg import
#' @import utils
#' @import BiocFileCache
#' @param hl hail object
#' @param retrieve_import_write logical(1) if TRUE, use hl.utils.get_1kg to retrieve data, otherwise acquire
#' a previously written zip file, either from a cache, or, if no file found in cache, from web, followed by caching
#' @param path_1kg_zip character(1) path to zip of MatrixTable, defaults to `osn_1kg_path()`.
#' @param folder character(1) destination of 1kg.mt as retrieved using hl.utils.get_1kg, import_vcf, write
#' @param cache a BiocFileCache-type cache
#' @note overwrite is permitted in the import_vcf.write event
#' @return "hail.matrixtable.MatrixTable" instance
#' @examples
#' hl <- hail_init()
#' mt <- get_1kg(hl)
#' mt
#' mt$rows()$select()$show(5L) # must use integer
#' annopath <- path_1kg_annotations()
#' tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
#' tab$describe()
#' tab$show(width = 100L)
#' @export
get_1kg <- function(hl, retrieve_import_write = FALSE, path_1kg_zip = osn_1kg_path(), folder = tempdir(), cache = BiocFileCache::BiocFileCache()) {
  proc <- basilisk::basiliskStart(bsklenv, testload = "hail") # avoid package-specific import
  on.exit(basilisk::basiliskStop(proc))
  basilisk::basiliskRun(proc, function(hl, folder) {
    if (retrieve_import_write) {
      hl$utils$get_1kg(folder)
      hl$import_vcf(paste0(folder, "/1kg.vcf.bgz"))$write(paste0(folder, "/1kg.mt"), overwrite = TRUE)
    } else {
      #       zf = system.file("extdata/1kg.zip", package="BiocHail")
      qout <- bfcquery(cache, "hail_tut_mt_zip")
      if (nrow(qout) == 0) { # download and populate cache
        dl <- try(bfcadd(cache,
          fpath = path_1kg_zip,
          rname = "hail_tut_mt_zip", rtype = "web", action = "copy",
          download = TRUE
        ))
        if (inherits(dl, "try-error")) stop("no hail_tut_mt in cache and cannot populate with download, try another method.")
        qout <- bfcquery(cache, "hail_tut_mt")
        stopifnot(nrow(qout) == 1L)
      }
      utils::unzip(cache[[qout$rid]], exdir = folder)
    }
    mt <- hl$read_matrix_table(paste0(folder, "/1kg.mt"))
    mt
  }, hl = hl, folder = folder)
}

#' generate path to installed annotations file
#' @note .txt file retrieved from extraction on `https://storage.googleapis.com/hail-1kg/tutorial_data.tar`
#' @return character(1) path to annotations
#' @examples
#' path_1kg_annotations()
#' @export
path_1kg_annotations <- function() {
  system.file("extdata/1kg_annotations.txt", package = "BiocHail")
}

# this test script introduced when vignette processing of
# these computations started to fail with too close to stack limit
# the computations work, just not in the context of vignette rendering as of 21 Aug 2023

test_that("gwas can be conducted", {

## ----getlib,message=FALSE-----------------------------------------------------
library(BiocHail)
library(ggplot2)


## ----get1, eval=TRUE----------------------------------------------------------
hl <- hail_init()
mt <- get_1kg(hl)
mt
print(mt$rows()$select()$show(5L)) # limited info


## r.mt.rows().select().show(5) # python chunk!


## r.mt.s.show(5)  # python chunk!


## ----getdim, eval=TRUE--------------------------------------------------------
mt$count()


## ----dodim, eval=TRUE---------------------------------------------------------
dim.hail.matrixtable.MatrixTable <- function(x) { 
  tmp <- x$count()
  c(tmp[[1]], tmp[[2]]) 
}
dim(mt)
ncol.hail.matrixtable.MatrixTable <- function(x) { 
 dim(x)[2]
}
nrow.hail.matrixtable.MatrixTable <- function(x) { 
 dim(x)[1]
}
nrow(mt)


## ----domo, eval=TRUE----------------------------------------------------------
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute=TRUE)$key_by("Sample")

tab$describe()
tab$show(width=100L)

mt2 = mt$annotate_cols(pheno = tab[mt$s])
mt = mt2

mt$col$describe()

## r.mt = r.mt.annotate_cols(pheno = r.tab[r.mt.s])   # python chunk!

## r.mt.col.describe()


## ----getsup, eval=TRUE--------------------------------------------------------
mt$aggregate_cols(hl$agg$counter(mt$pheno$SuperPopulation))


## ----lkcaf, eval=TRUE---------------------------------------------------------
uu <- mt$aggregate_cols(hl$agg$stats(mt$pheno$CaffeineConsumption))
names(uu)
uu$mean
uu$stdev


## from pprint import pprint  # python chunk!

## snp_counts = r.mt.aggregate_rows(r.hl.agg.counter(r.hl.Struct(ref=r.mt.alleles[0], alt=r.mt.alleles[1])))

alls = mt$alleles$collect()
al1 = sapply(alls, "[", 1)
al2 = sapply(alls, "[", 2)
snp_counts = mt$aggregate_rows(hl$agg$counter(hl$Struct(ref=al1, alt=al2)))
snp_counts

## pprint(snp_counts)


## ----dohist, eval=TRUE--------------------------------------------------------
p_hist <- mt$aggregate_entries(
     hl$expr$aggregators$hist(mt$DP, 0L, 30L, 30L))
names(p_hist)
length(p_hist$bin_edges)
length(p_hist$bin_freq)
midpts <- function(x) diff(x)/2+x[-length(x)]
dpdf <- data.frame(x=midpts(p_hist$bin_edges), y=p_hist$bin_freq)
ggplot(dpdf, aes(x=x,y=y)) + geom_col() + ggtitle("DP") + ylab("Frequency")


## ----lkagg, eval=TRUE---------------------------------------------------------
names(hl$expr$aggregators)


## ----update, eval=TRUE--------------------------------------------------------
mt <- hl$sample_qc(mt)

## r.mt.col.describe()  # python!


## ----lkcr, eval=TRUE----------------------------------------------------------
callrate <- mt$sample_qc$call_rate$collect()
graphics::hist(callrate)


## # python chunk!

## r.mt = r.mt.filter_cols((r.mt.sample_qc.dp_stats.mean >= 4) & (r.mt.sample_qc.call_rate >= 0.97))

mt$count_cols()
mt = mt$filter_cols((mt$sample_qc$dp_stats$mean >= 4 & mt$sample_qc$call_rate >= 0.97))
mt$count_cols()

## print('After filter, %d/284 samples remain.' % r.mt.count_cols())


## ab = r.mt.AD[1] / r.hl.sum(r.mt.AD)

## 

## filter_condition_ab = ((r.mt.GT.is_hom_ref() & (ab <= 0.1)) |

##                         (r.mt.GT.is_het() & (ab >= 0.25) & (ab <= 0.75)) |

##                         (r.mt.GT.is_hom_var() & (ab >= 0.9)))

## 

## fraction_filtered = r.mt.aggregate_entries(r.hl.agg.fraction(~filter_condition_ab))

## print(f'Filtering {fraction_filtered * 100:.2f}% entries out of downstream analysis.')

## r.mt = r.mt.filter_entries(filter_condition_ab)


## ----after, eval=TRUE---------------------------------------------------------
dim(mt)


## ----addvarqc, eval=TRUE------------------------------------------------------
mt = hl$variant_qc(mt)

## r.mt.row.describe()  #! python


## r.mt = r.mt.filter_rows(r.mt.variant_qc.AF[1] > 0.01)

## r.mt = r.mt.filter_rows(r.mt.variant_qc.p_value_hwe > 1e-6)

## r.mt.count()

gwas = hl$linear_regression_rows(y=mt$pheno$CaffeineConsumption,
  x=mt$GT$n_alt_alleles(), covariates=list(1.0))
allp = gwas$p_value$collect()
summary(allp)

expect_true( abs( mean(allp, na.rm=TRUE) - 0.306 ) < .001 )
})

## r.gwas = r.hl.linear_regression_rows(y=r.mt.pheno.CaffeineConsumption,

##                                  x=r.mt.GT.n_alt_alleles(),

##                                  covariates=[1.0])

## # r.pl = r.hl.plot.manhattan(r.gwas.p_value)

## # import bokeh

## # bokeh.plotting.show(r.pl)


## ----dolam, eval=TRUE---------------------------------------------------------
###pv = gwas$p_value$collect()
###x2 = stats::qchisq(1-pv,1)
###lam = stats::median(x2, na.rm=TRUE)/stats::qchisq(.5,1)
###lam
###
###
##### ----doqq, eval=TRUE----------------------------------------------------------
###qqplot(-log10(ppoints(length(pv))), -log10(pv), xlim=c(0,8), ylim=c(0,8),
###  ylab="-log10 p", xlab="expected")
###abline(0,1)
###
###
##### ----eval=FALSE---------------------------------------------------------------
##### locs <- gwas$locus$collect()
##### conts <- sapply(locs, function(x) x$contig)
##### pos <- sapply(locs, function(x) x$position)
##### library(igvR)
##### mytab <- data.frame(chr=as.character(conts), pos=pos, pval=pv)
##### gt <- GWASTrack("simp", mytab, chrom.col=1, pos.col=2, pval.col=3)
##### igv <- igvR()
##### setGenome(igv, "hg19")
##### displayTrack(igv, gt)
###
###
##### r.pcastuff = r.hl.hwe_normalized_pca(r.mt.GT)
###
##### r.mt = r.mt.annotate_cols(scores=r.pcastuff[1][r.mt.s].scores)
###
###
##### ----lkpcs, eval=TRUE---------------------------------------------------------
###sc <- pcastuff[[2]]$scores$collect()
###pc1 = sapply(sc, "[", 1)
###pc2 = sapply(sc, "[", 2)
###fac = mt$pheno$SuperPopulation$collect()
###myd = data.frame(pc1, pc2, pop=fac)
###library(ggplot2)
###ggplot(myd, aes(x=pc1, y=pc2, colour=factor(pop))) + geom_point()
###
###
##### r.gwas2 = r.hl.linear_regression_rows(
###
#####     y=r.mt.pheno.CaffeineConsumption,
###
#####     x=r.mt.GT.n_alt_alleles(),
###
#####     covariates=[1.0,r.mt.pheno.isFemale,r.mt.scores[0],
###
#####         r.mt.scores[1], r.mt.scores[2]])
###
###
##### ----dolam2, eval=TRUE--------------------------------------------------------
###pv = gwas2$p_value$collect()
###x2 = stats::qchisq(1-pv,1)
###lam = stats::median(x2, na.rm=TRUE)/stats::qchisq(.5,1)
###lam
###
###
##### ----doman, eval=TRUE---------------------------------------------------------
###locs <- gwas2$locus$collect()
###conts <- sapply(locs, function(x) x$contig)
###pos <- sapply(locs, function(x) x$position)
###mytab <- data.frame(chr=as.character(conts), pos=pos, pval=pv)
###ggplot(mytab[mytab$chr=="8",], aes(x=pos, y=-log10(pval))) + geom_point()
###
###
##### ----sessionInfo--------------------------------------------------------------
###sessionInfo()
###

library(BiocHail)

context("tutorial-related code works")

test_that("get_1kg succeeds", {
  hl = hail_init()
  x = get_1kg(hl) # would populate, pull from cache
  cc = x$count()
  expect_true(all(unlist(cc)==c(10961, 284)))
})

methods::setOldClass("hail.table.Table")

#' extract field names from hail.table.Table
#' @importFrom BiocGenerics colnames as.data.frame
#' @param x hail.table.Table instance
#' @param do.NULL ignored
#' @param prefix ignored
#' @return character vector
#' @examples
#' hl <- hail_init()
#' annopath <- path_1kg_annotations()
#' tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
#' colnames(tab)
#' @export
setMethod(
  "colnames", "hail.table.Table",
  function(x, do.NULL = TRUE, prefix = "col") {
    tf <- tempfile()
    on.exit(unlink(tf))
    zz <- reticulate::py_capture_output(x$head(1L)$export(tf)) # unpleasant, trying to find a better way
    names(read.delim(tf))
  }
)

#' S3 generic for get_key
#' @param x anything
#' @return typically a list
#' @examples
#' hl <- hail_init()
#' annopath <- path_1kg_annotations()
#' tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
#' get_key(tab)
#' @export
get_key <- function(x) UseMethod("get_key")

#' S3 method for get_key
#' @param x instance of hail.table.Table
#' @return a list with elements names (names of keys) and key_df (data.frame of key values, with column names)
#' @examples
#' hl <- hail_init()
#' annopath <- path_1kg_annotations()
#' tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
#' get_key(tab)
#' @export
get_key.hail.table.Table <- function(x) {
  tf <- paste0(tempfile(), ".csv")
  zz <- reticulate::py_capture_output(x$key$export(tf))
  ans <- read.csv(tf)
  list(names = names(ans), key_df = ans)
}

#' acquire row names of a Hail Table, assuming key has been set
#' @importFrom BiocGenerics rownames
#' @return character()
#' @param x instance of hail.table.Table
#' @param do.NULL not used
#' @param prefix not used
#' @note To try example, run `example("rownames,hail.table.Table-method")`
#' @return character vector
#' @examples
#' hl <- hail_init()
#' annopath <- path_1kg_annotations()
#' tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
#' rt <- rownames(tab)
#' length(rt)
#' head(rt)
#' @export
setMethod(
  "rownames", "hail.table.Table",
  function(x, do.NULL = TRUE, prefix = "row") {
    kk <- get_key(x)
    x$key$get(kk[[1]])$collect()
  }
)

#' initialize ukbb
#' @import basilisk
#' @note ukbb module may be passed around
#' @return python module reference
#' @examples
#' ukbb <- ukbb_init()
#' names(ukbb)
#' @export
ukbb_init <- function() {
  proc <- basilisk::basiliskStart(bsklenv, testload = "hail") # avoid package-specific import
  # on.exit(basilisk::basiliskStop(proc))
  basilisk::basiliskRun(proc, function() {
    ukbb <- reticulate::import("ukbb_pan_ancestry")
    ukbb
  })
}

#' initialize hail
#' @import basilisk
#' @note hail object may be passed around
#' @return python reference to hail module
#' @examples
#' hc <- hail_init_simple()
#' hc
#' @export
hail_init_simple = function() {
 proc = basilisk::basiliskStart(bsklenv, testload="hail") # avoid package-specific import
 #on.exit(basilisk::basiliskStop(proc))
 basilisk::basiliskRun(proc, function() {
     hl = reticulate::import("hail") 
     lk = try(hl$init(idempotent=TRUE))
     if (inherits(lk, "try-error")) {
        message("could not initialize ... already initialized?")
        }
     if (length(names(hl))==0) stop("invalid hail module, you may need to reboot...")
     hl
   })
}

#' stop hail
#' @param hl a hail object produced by hail_init()
#' @return result of stop() method for Hail module
#' @examples
#' hail_stop
#' @export
hail_stop = function(hl) hl$stop()

#init(sc=None, app_name=None, master=None, local='local[*]', log=None, quiet=False, append=False, min_block_size=0, branching_factor=50, tmp_dir=None, default_reference='GRCh37', idempotent=False, global_seed=None, spark_conf=None, skip_logging_configuration=False, local_tmpdir=None, _optimizer_iterations=None, *, backend=None, driver_cores=None, driver_memory=None, worker_cores=None, worker_memory=None, gcs_requester_pays_configuration: Union[str, Tuple[str, List[str]], NoneType] = None)

#    >>> hl.init(
#    ...     gcs_requester_pays_configuration=('my-project', ['bucket_of_fish', 'bucket_of_eels'])
#    ... )  # doctest: +SKIP
    



#' initialize hail, using more options
#' @param quiet logical(1) defaults to FALSE
#' @param min_block_size integer(1) defaults to 0L 
#' @param branching_factor integer(1) defaults to 50L
#' @param default_reference character(1) defaults to "GRCh37", for compatibility with earlier `hail_init`
#' @param global_seed integer(1) defaults to 1234L 
#' @param log character(1) target folder for logging, defaults to tempfile()
#' @param spark_conf list, defaults to NULL
#' @param gcs_requester_pays_configuration list, defaults to NULL
#' @note hail object may be passed around.  See hail documentation for details on all args.
#' @return python reference to hail module
#' @examples
#' proj = Sys.getenv("GOOGLE_PROJECT")
#' buck = Sys.getenv("GCS_BUCKET")
#' if (nchar(buck)>0) {
#'   # conf = list(proj, c(buck)) doesn't seem to generate tuple[str,Sequence[str]]
#'   hl <- hail_init()   #gcs_requester_pays_configuration=conf)
#'   hl$default_reference()
#' }  
#' @export
hail_init = function(quiet=FALSE, min_block_size=0L, branching_factor=50L,
   default_reference="GRCh37", global_seed=1234L, log = tempfile(), spark_conf=NULL,
   gcs_requester_pays_configuration = NULL) {
 proc = basilisk::basiliskStart(bsklenv, testload="hail") # avoid package-specific import
 #on.exit(basilisk::basiliskStop(proc))
 basilisk::basiliskRun(proc, function(quiet, min_block_size, branching_factor,
		default_reference, global_seed, log, spark_conf, gcs_requester_pays_configuration) {
     hl = reticulate::import("hail") 
     lk = try(hl$init(idempotent=TRUE, quiet=quiet, min_block_size=min_block_size,
               branching_factor=branching_factor, default_reference=default_reference,
               global_seed=global_seed, log=log, gcs_requester_pays_configuration=
                  gcs_requester_pays_configuration))
     if (inherits(lk, "try-error")) {
        message("could not initialize ... already initialized?")
        }
     hl
   }, quiet=quiet, min_block_size=min_block_size, branching_factor=branching_factor,
        default_reference=default_reference, global_seed=global_seed, log=log,
         spark_conf=spark_conf, gcs_requester_pays_configuration=
         gcs_requester_pays_configuration  )
}

#' bare interface to hail using reticulate
#' @importFrom reticulate import
#' @note `/home/jupyter/.local/share/r-miniconda/envs/r-reticulate/bin/pip3 install...` is 
#' used to ensure that reticulate's python ecosystem is what we want
#' @return python reference to hail module
#' @examples
#' # assumes terra
#' if (nchar(Sys.getenv("WORKSPACE_NAMESPACE"))>0) {
#'   hl = bare_hail()
#'   hl$init(idempotent=TRUE, spark_conf=list(
#'       'spark.hadoop.fs.gs.requester.pays.mode'= 'CUSTOM',
#'       'spark.hadoop.fs.gs.requester.pays.buckets'= 'ukb-diverse-pops-public',
#'       'spark.hadoop.fs.gs.requester.pays.project.id'= Sys.getenv("GOOGLE_PROJECT")))
#'   hl$read_matrix_table('gs://ukb-diverse-pops-public/sumstats_release/results_full.mt')$describe()
#'   \dontrun{
#'   # this is supposed to get us some LD data but xx.shape fails, issue filed
#'     hli = reticulate::import("hail.linalg")
#'     upa = reticulate::import("ukbb_pan_ancestry")
#'     xx = hli$BlockMatrix$read(upa$get_ld_matrix_path('AFR'))
#'    }
#'   }
#' @export
bare_hail = function() {
  reticulate::import("hail")
}

library(BiocHail)

context("hail_init code works")

test_that("hail_init succeeds and is idempotent", {
   hl = hail_init()
   expect_true(substr(hl$version(),1,7)=="0.2.135")
   hl = hail_init()
   expect_true(substr(hl$version(),1,7)=="0.2.135")
})

#' S3 support
#' @param x entity coercible to data.frame
#' @param row.names character or NULL
#' @param optional logical
#' @param \dots any args
#' @return data.frame
#' @examples
#' hl <- hail_init()
#' annopath <- path_1kg_annotations()
#' tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
#' as.data.frame(tab$head(3L))
#' @export
as.data.frame <- function(x, row.names = NULL, optional = FALSE, ...) UseMethod("as.data.frame")

#' S3 support
#' @param x entity coercible to data.frame
#' @param row.names character or NULL
#' @param optional logical
#' @param \dots any args
#' @return data.frame
#' @examples
#' hl <- hail_init()
#' annopath <- path_1kg_annotations()
#' tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
#' as.data.frame(tab$head(3L))
#' @method as.data.frame default
as.data.frame.default <- base::as.data.frame

mkrow <- function(str, cn) lapply(cn, function(x) str$get(x))

methods::setOldClass("hail.table.Table")

#' convert hail.table.Table to R data frame
#' @note only use on small table because collect is used
#' @param x instance of "hail.table.Table"
#' @param row.names not used
#' @param optional not used
#' @param \dots not used
#' @return data.frame
#' @examples
#' hl <- hail_init()
#' annopath <- path_1kg_annotations()
#' tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
#' as.data.frame(tab$head(3L))
#' @export
as.data.frame.hail.table.Table <- function(x, row.names = NULL, optional = FALSE, ...) {
  cx <- x$collect()
  cn <- colnames(x)
  tmpl <- lapply(cx, function(z) {
    tmp <- data.frame(mkrow(z, cn))
    names(tmp) <- paste0("x", seq_len(ncol(tmp)))
    tmp
  })
  tmpdf <- do.call(rbind, tmpl)
  names(tmpdf) <- cn
  tmpdf
}

#' HWE-normalized PCA scores for 3202 thousand-genomes samples genotyped with the
#' telomere-to-telomere reference
#' @note The genotypes are from a 1% sample of loci on chr17
#' @docType data
#' @usage data("pcs_38k")
#' @return data.frame
#' @examples
#' data(pcs_38k)
#' dim(pcs_38k)
#' @format data.frame
"pcs_38k"

#' HWE-normalized PCA scores for 3202 thousand-genomes samples genotyped with the
#' telomere-to-telomere reference
#' @note The genotypes are from a 5% sample of loci on chr17
#' @docType data
#' @usage data("pcs_191k")
#' @return data.frame
#' @examples
#' data(pcs_191k)
#' dim(pcs_191k)
#' @format data.frame
"pcs_191k"

#' data.frame with metadata about 3202 samples genotyped against T2T reference
#' @note Source: index files described at `https://www.internationalgenome.org/data-portal/data-collection/30x-grch38`
#' @docType data
#' @usage data("kg_3202")
#' @return data.frame
#' @examples
#' data(kg_3202)
#' dim(kg_3202)
#' @format data.frame
"kg_3202"

#' Open Storage Network path to a zip of hail MatrixTable with a small subset of UKBB summary statistics as of 12/25/2022
#' @return character(1) path to zip
#' @examples
#' osn_ukbb_sumst10k_path()
#' @export
osn_ukbb_sumst10k_path <- function() "https://bir190004-bucket01.mghp.osn.xsede.org/BiocUKBBData/ukbb_sumst_10kloc.zip"

#' interface to a small subset of UKBB summary stats in MatrixTable format
#' @import utils
#' @import BiocFileCache
#' @param hl hail object
#' @param folder character(1) destination of 1kg.mt as retrieved using hl.utils.get_1kg, import_vcf, write
#' @param cache a BiocFileCache-type cache
#' @param timeout.ukbb numeric(1) defaults to 3600 for timeout setting in `options()`; option value is reset on exit
#' @return "hail.matrixtable.MatrixTable" instance
#' @note The loci were selected using a .000345% random sample
#' of over 28 million loci recorded in the UKBB pan-ancestry
#' record.  The sample is made available a) to assess some issues with data volume (the full
#' resource is about 12.78 TB according to [this doc](https://pan-dev.ukbb.broadinstitute.org/docs/hail-format/index.html), and b) to provide full information on the scope of phenotypes and populations available.
#' @note This function will unzip 5GB of MatrixTable data.  It may be desirable to cache the
#' unzipped image to a persistent location.  If this has been done and the environment variable
#' `HAIL_UKBB_SUMSTAT_10K_PATH` has been set to this location, this function will use
#' the MatrixTable content found there.
#' @examples
#' hl <- hail_init()
#' ss <- get_ukbb_sumstat_10kloci_mt(hl)
#' # consider saving the unzipped image and recaching
#' ss$count()
#' @export
get_ukbb_sumstat_10kloci_mt <- function(hl, folder = tempdir(), cache = BiocFileCache::BiocFileCache(), timeout.ukbb = 3600) {
  proc <- basilisk::basiliskStart(bsklenv, testload = "hail") # avoid package-specific import
  on.exit(basilisk::basiliskStop(proc))
  basilisk::basiliskRun(proc, function(hl, folder) {
    #
    # check for local archive and use it if available
    #
    lklocal <- Sys.getenv("HAIL_UKBB_SUMSTAT_10K_PATH")
    if (nchar(lklocal) > 0) {
      stopifnot(basename(lklocal)=="ukbb_sumstat_10k_loci.mt") # the cached zip must have been uncached here
      #
      # prepare for long download
      #
      return(hl$read_matrix_table(lklocal))
    }
    ot <- options()$timeout
    on.exit(options(timeout = ot))
    options(timeout = timeout.ukbb)
    #
    # check cache for zip, populate if not present
    #
    qout <- bfcquery(cache, "ukbb_sumst_10kloc.zip")
    if (nrow(qout) == 0) { # download and populate cache
      dl <- try(bfcadd(cache,
        fpath = osn_ukbb_sumst10k_path(),
        rname = "ukbb_sumst_10kloc.zip", rtype = "web", action = "copy",
        download = TRUE
      ))
      if (inherits(dl, "try-error")) stop("no ukbb_sumst in cache and cannot populate with download, try another method.")
      qout <- bfcquery(cache, "ukbb_sumst_10kloc.zip")
      stopifnot(nrow(qout) == 1L)
    }
    #
    # unzip to folder
    #
    utils::unzip(cache[[qout$rid]], exdir = folder)
    mt <- hl$read_matrix_table(paste0(folder, "/ukbb_sumstat_10k_loci.mt"))
    mt
  }, hl = hl, folder = folder)
}

#' s3 support
#' @param .data instance of hail.table.Table
#' @param \dots should include named components `filter` which is a logical vector
#' with same number of rows as `.data`, `hl`, a reference to a hail environment (Module),
#' and `placeholder` an arbitrary character(1)
#' @param .by not used
#' @param .preserve not used
#' @return filtered hail.table.Table reference
#' @examples
#' hl <- hail_init()
#' annopath <- path_1kg_annotations()
#' tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
#' pick <- rep(FALSE, 3500)
#' pick[seq_len(10)] <- TRUE
#' ft <- filter(tab, filter = pick, hl = hl)
#' ft$count()
#' ft$head(2L)$collect()
#' @export
filter <- function(.data, ..., .by = NULL, .preserve = FALSE) UseMethod("filter")


#' filter rows of a hail Table
#' @importFrom dplyr filter
#' @import methods
#' @return filtered hail.table.Table reference
#' @note writes one line of table to disk to retrieve field names
#' @param .data instance of hail.table.Table
#' @param \dots should include named components `filter` which is a logical vector
#' with same number of rows as `.data`, `hl`, a reference to a hail environment (Module),
#' and `placeholder` an arbitrary character(1)
#' @param .by not used
#' @param .preserve not used
#' @note FIXME: uses disk because I don't know how to create a BooleanExpression except by importing.
#' @examples
#' hl <- hail_init()
#' annopath <- path_1kg_annotations()
#' tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
#' pick <- rep(FALSE, 3500)
#' pick[seq_len(10)] <- TRUE
#' ft <- filter(tab, filter = pick, hl = hl)
#' ft$count()
#' ft$head(2L)$collect()
#' @export
filter.hail.table.Table <- function(.data, ..., .by = NULL, .preserve = FALSE) {
  input <- list(...)
  nms <- names(input)
  stopifnot(all(c("filter", "hl") %in% nms))
  hl <- input$hl
  filter <- input$filter
  placeholder <- "tmpf"
  if (!is.null(input$placeholder)) placeholder <- input$placeholder
  x <- .data
  if (!inherits(filter, "hail.expr.expressions.typed_expressions.BooleanExpression")) {
    # must build such
    stopifnot(is(filter, "logical"))
    nrec <- x$count()
    if ((nf <- length(filter)) != nrec) {
      stop(sprintf(
        "filter must have length %d; got %d.\n",
        nrec, nf
      ))
    }
    kk <- get_key(x)
    tmp <- data.frame(k = kk[[2]], filter)
    names(tmp) <- c(kk[[1]], placeholder) # get key name from table
    tf <- paste0(tempfile(), ".csv")
    write.csv(tmp, tf, row.names = FALSE, quote = FALSE)
    tlist <- list(hl$tstr, hl$tbool)
    names(tlist) <- c(kk[[1]], placeholder)
    y <- hl$import_table(tf, types = tlist, delimiter = ",")
    x <- x$join(y$key_by(kk[[1]]))
    filter <- x$tmpf
  }
  fil <- x$filter(filter)
  fil
}

#' top-level annotation harvesting from columns of summary statistics MatrixTable
#' @param x a Struct returned from mt$cols()$collect() -- which can be slow
#' @param toget character() vector of field names to retrieve
#' @note python None are transformed to NA
#' @return 1-row data.frame
top2df <- function(x, toget = c("trait_type", "phenocode", "description", "modifier", "coding_description")) {
  # must deal with python 'None'
  z <- data.frame(lapply(toget, function(w) {
    chk <- x[[w]]
    if (is.null(chk)) chk <- NA
    chk
  }))
  names(z) <- toget
  z
}

#' pheno_data component harvesting from columns of summary stats MatrixTable
#' @param m Struct returned from mt$cols()$collect()
#' @param section numeric(1) element of pheno_data list to be transformed to data.frame
#' @param toget character() vector of field names to retrieve
#' @param verbose logical(1) if TRUE (NOT default) will message that there are multiple
#' `pheno_data` components returned
#' @note applies top2df to the pheno_data component of input
#' @return 1 row data.frame
pheno_data_sec_2df <- function(m, section = 1,
                               toget = c("n_cases", "n_controls", "heritability", "pop"), verbose = FALSE) {
  pd <- m$pheno_data
  npd <- length(pd)
  if (npd > 1) {
    if (section > npd) stop("requested section number in excess of available pheno_data elements returned")
    if (verbose) {
      message(sprintf("there are %d pheno_data components, selecting component %d", npd, section))
    }
  }
  z <- pd[[section]]
  top2df(z, toget = toget)
}

#' pheno_data component harvesting from columns of summary stats MatrixTable
#' allowing for info on multiple populations in the pheno_data component
#' @param x Struct - a single element of the list returned by mt$cols()$collect()
#' @param top2get character() vector of general fields to retrieve
#' @param pheno2get character() vector of fields to be retrieved for each subpopulation
#' @return data.frame
#' @examples
#' # following are too time-consuming but can be of interest
#' # if (nchar(Sys.getenv("HAIL_UKBB_SUMSTAT_10K_PATH"))>0) {
#' #  hl = hail_init()
#' #  ss = get_ukbb_sumstat_10kloci_mt(hl)
#' #  sscol = ss$cols()$collect() # may take a bit of time
#' #  print(length(sscol))
#' #  multipop_df(sscol[[1]])
#' # }
#' #
#' \donttest{
#' # if (nchar(Sys.getenv("HAIL_UKBB_SUMSTAT_10K_PATH"))>0) {
#' # # to get an overview of all phenotype-cohort combinations in a searchable table
#' # mmm = lapply(sscol, multipop_df )
#' # mymy = do.call(rbind, mmm) # over 16k rows
#' # DT::datatable(mymy)
#' # }
#' #
#' }
#' # this runs quickly and is demonstrative
#' hl <- hail_init()
#' litzip <- system.file("extdata", "myss2.zip", package = "BiocHail")
#' td <- tempdir()
#' unzip(litzip, exdir = td)
#' ntab <- hl$read_matrix_table(paste0(td, "/myss2.mt"))
#' ntab$describe()
#' nt2 <- ntab$col$collect()
#' multipop_df(nt2[[1]]) # must select one element
#' @export
multipop_df <- function(x, top2get = c(
                          "trait_type", "phenocode",
                          "description", "modifier", "coding_description", "coding"
                        ),
                        pheno2get = c("n_cases", "n_controls", "heritability", "pop")) {
  nph <- length(x$pheno_data)
  ini <- top2df(x, toget = top2get)
  multi <- do.call(rbind, lapply(seq_len(nph), function(z) {
    pheno_data_sec_2df(x,
      section = z,
      toget = pheno2get
    )
  }))
  cbind(ini, multi)
}

#!/usr/bin/env bash

# 12/27/2022 will produce R 4.2.2 in docker ubuntu:22.04 in 2 minutes

# added
apt update
apt install --yes ca-certificates
DEBIAN_FRONTEND=noninteractive apt-get install -y tzdata

apt install --yes --no-install-recommends wget  	# to add the key
wget -q -O- https://eddelbuettel.github.io/r2u/assets/dirk_eddelbuettel_key.asc \
    | tee -a /etc/apt/trusted.gpg.d/cranapt_key.asc

# USE FOCAL

#echo "deb [arch=amd64] https://dirk.eddelbuettel.com/cranapt focal main" \
#    > /etc/apt/sources.list.d/cranapt.list
#apt update
# or use the mirror at the University of Illinois Urbana-Champaign:

echo "deb [arch=amd64] https://r2u.stat.illinois.edu/ubuntu focal main" \
    > /etc/apt/sources.list.d/cranapt.list
apt update

# (In either example, replace focal with focal for use with Ubuntu 22.04.)

#Third, and optionally, if you do not yet have the current R version, run these two lines (or use the standard CRAN repo setup)
#

wget -q -O- https://cloud.r-project.org/bin/linux/ubuntu/marutter_pubkey.asc \
    | tee -a /etc/apt/trusted.gpg.d/cran_ubuntu_key.asc
echo "deb [arch=amd64] https://cloud.r-project.org/bin/linux/ubuntu focal-cran40/" \
    > /etc/apt/sources.list.d/cran-ubuntu.list
apt update
apt install -y r-base-core

#!/usr/bin/env bash

# This script allows us to have R 4.2.2 in a Hail/Spark cluster on terra
# At this time, BiocHail's basilisk-based hail_init(2) won't interface to gs:// storage,
# but pure reticulate interface to hail does

# This script builds R from source as it has been difficult to find debian packaging
# yielding R 4.2

conda update -n base -c defaults conda
apt -y upgrade
apt-get update
apt-get install -y r-base-dev
apt-get install -y subversion
apt-get install -y rsync
apt-get install -y libcurl4-gnutls-dev
apt-get install -y vim
apt-get install openjdk-8-jdk
svn co https://svn.r-project.org/R/branches/R-4-2-branch R-4-2-src
cd R-4-2-src
cd tools
./rsync-recommended
cd ..  
./configure --enable-R-shlib --with-x=no
make -j 2
make install
/usr/local/bin/R CMD javareconf
/usr/local/bin/R -e "options(repos=c(CRAN = 'https://cloud.r-project.org'));install.packages(c('remotes', 'BiocManager'))"
export PIP_USER=false
/usr/local/bin/R -e "BiocManager::install('vjcitn/BiocHail')"
#wget https://www.openssl.org/source/openssl-3.0.7.tar.gz
#tar zxf openssl-3.0.7.tar.gz
#cd openssl-3.0.7
#./config --prefix=/usr/local/ssl --openssldir=/usr/local/ssl shared zlib
#make
#make install
#echo /usr/local/ssl/lib >> /etc/ld.so.conf.d/openssl-3.0.7.ld.conf
#ldconfig
#/usr/local/bin/R -e "library(BiocHail); example(get_1kg)"

---
title: "03 Working with UK Biobank summary statistics"
author: "Vincent J. Carey, stvjc at channing.harvard.edu"
date: "`r format(Sys.time(), '%B %d, %Y')`"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{03 Working with UK Biobank summary statistics}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document:
    highlight: pygments
    number_sections: yes
    theme: united
    toc: yes
---

# Overview

In this document we illustrate some approaches to
working with UK Biobank summary statistics.
Be sure that that the python module `ukbb_pan_ancestry`
has been installed where reticulate can find it.
(We don't use basilisk as of 12/24/2022 because of
issues in the terra spark cluster.)

```{r setup,message=FALSE, eval=FALSE}
library(BiocHail)
```

If the above command indicates that BiocHail is not available,
see the Installation section near the end of this document.

## Initialization and description

### Standalone

We have produced a representation of summary
statistics for a sample of 9888 loci.  This 5GB resource
can be retrieved and cached with the following code:
```{r getukbb, eval=FALSE}
hl = hail_init()
ss = get_ukbb_sumstat_10kloci_mt(hl) # can take about a minute to unzip 5GB
ss$count()   # but if a persistent MatrixTable is at the location given
             # by env var HAIL_UKBB_SUMSTAT_10K_PATH it goes quickly
```
To get a description of available content, we need a python chunk:
```{python lkdesc, eval=FALSE}
r.ss.describe()
```

### Terra
Here's a basic description of the summary stats table, with code that
works in terra.bio:
```{r lkba, eval=FALSE}
hl = bare_hail()
hl$init(idempotent=TRUE, spark_conf=list(
  'spark.hadoop.fs.gs.requester.pays.mode'= 'CUSTOM',
  'spark.hadoop.fs.gs.requester.pays.buckets'= 'ukb-diverse-pops-public',
  'spark.hadoop.fs.gs.requester.pays.project.id'= Sys.getenv("GOOGLE_PROJECT")))
```
We need to use a python chunk to get the output, using gs:// storage references.
```{python lkrhl, eval=FALSE}
r.hl.read_matrix_table('gs://ukb-diverse-pops-public/sumstats_release/results_full.mt').describe()
```

## Exploring the subset

### Metadata on study phenotypes

We'll collect the column metadata to start to understand details
of annotation of phenotypes.
```{r lkss, eval=FALSE}
sscol = ss$cols()$collect() # OK because we are just working over column metadata
ss1 = sscol[[1]]
names(ss1)
ss1$get("phenocode")
ss1$get("description")
```

We've introduced a function that extracts selected fields for a given
phenotype, that accommodates availability of results for specific populations.

```{r lkmul, eval=FALSE}
multipop_df(ss1)
```
This can be iterated over all the elements of `sscol` to produce
a comprehensive searchable table.  Here's a small illustration:
```{r dodt, message=FALSE, eval=FALSE}
library(DT)
ndf = do.call(rbind, lapply(sscol[1:10], multipop_df))
datatable(ndf)
```

### Metadata on loci

We'll trim the material to be worked with by sampling both rows and columns.
(2023.01.08: In future revisions we will be able to control the seed for random sampling.)

```{r dotrim, eval=FALSE}
sss = ss$sample_rows(.01)$sample_cols(.01)
sss$count()
```

Row metadata are simple to collect:
```{r getrm, eval=FALSE}
rss = sss$rows()$collect()
rss1 = rss[[1]]
names(rss1)
names(rss1$locus)
rss1$locus$contig
sapply(c("contig", "position"), function(x) rss1$locus[[x]])
```

### Summary statistics

The summary statistics themselves reside in entries of the MatrixTable.  This
can be expensive to collect and so filtering methods beyond random sampling
must be mastered.  But here is a basic view.

```{r getent, eval=FALSE}
sse = sss$entries()$collect()
length(sse)
names(sse[[1]])
sse1 = sse[[1]]
```

The `summary_stats` component has the association p-values -- log10 transformed?
```{r lkss11, eval=FALSE}
length(sse1$summary_stats)
names(sse1$summary_stats[[1]])
sse1$summary_stats[[1]]$Pvalue
```

# Exercises

## Infrastructure

- Define an interface to this subset of 10k loci that supports queries like
    - has disease x been studied in UK Biobank?
    - how many phenotypes have been studied in K populations, K=2, 3, ...?
    - how consistent is the annotation -- are numbers of controls and cases always
recorded?
- Do the UK Biobank portals produce information to resolve these questions?
    - if so, what are the API calls to obtain answers?
    - if not, what is missing from the portals to allow answers to be obtained?

## Substantive

- List the genes that are near the loci collected in the 10k random sample 
- (Hard) What is the most significant finding (position, phenotype, population) in the 10k subset?

# Installing `BiocHail`

`BiocHail` should be installed as follows:

```{r, eval = FALSE}
if (!require("BiocManager"))
    install.packages("BiocManager")
BiocManager::install("BiocHail")
```


# SessionInfo {-}

```{r sessionInfo}
sessionInfo()
```

---
title: "01 BiocHail -- GWAS tutorial from hail.is"
author: "Vincent J. Carey, stvjc at channing.harvard.edu"
date: "`r format(Sys.time(), '%B %d, %Y')`"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{01 BiocHail -- GWAS tutorial}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document:
    highlight: pygments
    number_sections: yes
    theme: united
    toc: yes
---

# Introduction

This document explores using Hail 0.2 with R
via basilisk.

The computations follow the
[GWAS tutorial](https://hail.is/docs/0.2/tutorials/01-genome-wide-association-study.html) 
in the hail documentation.  We won't do all the computations
there, and we add some material dealing with R-python
interfacing.  We'll note that the actual computations on
large data are done in Spark, but we don't interact directly
with Spark at all in this document.

Most of the computations are done via reticulate
calls to python; the access to the hail environment
is through basilisk.  We also take advantage of
R markdown's capacity to execute python code directly.
If an R chunk computes `x`, a python chunk can refer
to it as `r.x`.  If a python chunk computes `r.x`,
an R chunk can refer to this value as `x`.

# Installing `BiocHail`

`BiocHail` should be installed as follows:

```{r, eval = FALSE}
if (!require("BiocManager"))
    install.packages("BiocManager")
BiocManager::install("BiocHail")
```

As of 1.0.0, a JDK for Java version `<=` 11.0 is necessary
to use the version of Hail that is installed with the package.
This package should be usable on MacOS with suitable java
support.  If Java version `>=` 8.x is used, warnings from
Apache Spark may be observed.  To the best of our knowledge
the conditions to which the warnings pertain do not affect program performance.


# Acquire a slice of the 1000 genomes genotypes and annotations

In this section we import the 1000 genomes VCF slice
distributed by the hail project.  `hail_init` uses basilisk,
which ensures that a specific version of hail and
its dependencies are available in an isolated virtual environment.

```{r getlib,message=FALSE}
library(BiocHail)
library(ggplot2)
```

## Initialization, data acquisition, rendering

Here is a curiosity of R-hail interaction.  Note that
the following chunk computes `mt`, a MatrixTable
representation of 1000 genomes data, but our
attempt to print it in markdown fails.
```{r get1, eval=FALSE}
hl <- hail_init()
mt <- get_1kg(hl)
mt
print(mt$rows()$select()$show(5L)) # limited info
```

We can use the python syntax in a `python` R markdown
chunk to see what we want.  We use prefix `r.` to
find references defined in our R session (compiling
the vignette).
```{python abc, eval=FALSE}
r.mt.rows().select().show(5) # python chunk!
```

The sample IDs:
```{python ch2, eval=FALSE}
r.mt.s.show(5)  # python chunk!
```

## Helper functions

Some methods return data
immediately useful in R.
```{r getdim, eval=FALSE}
mt$count()
```
We can thus define a function `dim` to behave with
hail MatrixTable instances
in a familiar way, along with some others.
```{r dodim, eval=FALSE}
dim.hail.matrixtable.MatrixTable <- function(x) { 
  tmp <- x$count()
  c(tmp[[1]], tmp[[2]]) 
}
dim(mt)
ncol.hail.matrixtable.MatrixTable <- function(x) { 
 dim(x)[2]
}
nrow.hail.matrixtable.MatrixTable <- function(x) { 
 dim(x)[1]
}
nrow(mt)
```
These can be useful on their own, or when calling python methods.

## Acquiring `column fields`

```{r domo, eval=FALSE}
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute=TRUE)$key_by("Sample")
```
```{python ch3, eval=FALSE}
r.tab.describe()# python chunk!
r.tab.show(width=100)
```

## Adding the sample annotation to the MatrixTable; aggregation

We combine the `tab` defined above, with the MatrixTable instance,
using python code reaching to R via `r.`.
```{python ch4, eval=FALSE}
r.mt = r.mt.annotate_cols(pheno = r.tab[r.mt.s])   # python chunk! 
r.mt.col.describe()
```

Aggregation methods can be used to obtain contingency tables or descriptive statistics.

First, we get the frequencies of superpopulation membership:
```{r getsup, eval=FALSE}
mt$aggregate_cols(hl$agg$counter(mt$pheno$SuperPopulation))
```

Then statistics on caffeine consumption:
```{r lkcaf, eval=FALSE}
uu <- mt$aggregate_cols(hl$agg$stats(mt$pheno$CaffeineConsumption))
names(uu)
uu$mean
uu$stdev
```

# Working with variants; quality assessment

The significance of the aggregation functions is that the computations
are performed by Spark, on potentially huge distributed data structures.

Now we aggregate over rows (SNPs).  We'll use python directly:
```{python ch5, eval=FALSE}
from pprint import pprint  # python chunk!
snp_counts = r.mt.aggregate_rows(r.hl.agg.counter(r.hl.Struct(ref=r.mt.alleles[0], alt=r.mt.alleles[1])))
pprint(snp_counts)
```

## A histogram of read depths

Hail uses the concept of 'entries' for matrix elements, and each 'entry'
is a 'struct' with potentially many fields.

Here we'll use R to compute a histogram of sequencing depths
over all samples and variants.

```{r dohist, eval=FALSE}
p_hist <- mt$aggregate_entries(
     hl$expr$aggregators$hist(mt$DP, 0L, 30L, 30L))
names(p_hist)
length(p_hist$bin_edges)
length(p_hist$bin_freq)
midpts <- function(x) diff(x)/2+x[-length(x)]
dpdf <- data.frame(x=midpts(p_hist$bin_edges), y=p_hist$bin_freq)
ggplot(dpdf, aes(x=x,y=y)) + geom_col() + ggtitle("DP") + ylab("Frequency")
```

An exercise: produce a function `mt_hist` that produces
a histogram of measures from any of the relevant
VCF components of a MatrixTable instance.

Note also all the aggregators available:
```{r lkagg, eval=FALSE}
names(hl$expr$aggregators)
```

We'd also note that hail has a [direct interface to ggplot2](https://hail.is/docs/0.2/tutorials/09-ggplot.html).

## Quality summaries

A high-level function adds quality metrics to the MatrixTable.

```{r update, eval=FALSE}
mt <- hl$sample_qc(mt)
```{python ch6, eval=FALSE}
r.mt.col.describe()  # python!
```

The call rate histogram is given by:
```{r lkcr, eval=FALSE}
callrate <- mt$sample_qc$call_rate$collect()
graphics::hist(callrate)
```

## Filtering

### Sample quality

We'll use the python code given for filtering, in which per-sample
mean read depth and call rate are must exceed (arbitrarily chosen) thresholds.

```{python dofilt, eval=FALSE}
# python chunk!
r.mt = r.mt.filter_cols((r.mt.sample_qc.dp_stats.mean >= 4) & (r.mt.sample_qc.call_rate >= 0.97))
print('After filter, %d/284 samples remain.' % r.mt.count_cols())
```

### Genotype quality

Again we use the python code for filtering according to

- relative purity of reads underlying homozygous reference or alt calls
- good balance of ref/alt counts for het calls

```{python gtfilt, eval=FALSE}
ab = r.mt.AD[1] / r.hl.sum(r.mt.AD)

filter_condition_ab = ((r.mt.GT.is_hom_ref() & (ab <= 0.1)) |
                        (r.mt.GT.is_het() & (ab >= 0.25) & (ab <= 0.75)) |
                        (r.mt.GT.is_hom_var() & (ab >= 0.9)))

fraction_filtered = r.mt.aggregate_entries(r.hl.agg.fraction(~filter_condition_ab))
print(f'Filtering {fraction_filtered * 100:.2f}% entries out of downstream analysis.')
r.mt = r.mt.filter_entries(filter_condition_ab)
```
Note that filtering _entries_ does not change MatrixTable shape.

```{r after, eval=FALSE}
dim(mt)
```

### Variant characteristics

Allele frequencies, Hardy-Weinberg equilibrium test results
and other summaries are obtained using the `variant_qc` function.
```{r addvarqc, eval=FALSE}
mt = hl$variant_qc(mt)
```
```{python lkvarqc, eval=FALSE}
r.mt.row.describe()  #! python
```

# GWAS execution

A built-in procedure for testing for association between the
(simulated) caffeine consumption measure and genotype
will be used.

The following commands eliminate variants with minor allele frequency
less than 0.01, along with those with small $p$-values in
tests of Hardy-Weinberg equilibrium.

```{python dofilts, eval=FALSE}
r.mt = r.mt.filter_rows(r.mt.variant_qc.AF[1] > 0.01)
r.mt = r.mt.filter_rows(r.mt.variant_qc.p_value_hwe > 1e-6)
r.mt.count()
```

## Association test for quantitative response

Now we perform a naive test of association.  The Manhattan
plot generated by hail can be displayed for interaction
using bokeh.  We comment this out for now; it should
be possible to embed the bokeh display in this document but the details
are not ready-to-hand.
```{python naivet, eval=FALSE}
r.gwas = r.hl.linear_regression_rows(y=r.mt.pheno.CaffeineConsumption,
                                 x=r.mt.GT.n_alt_alleles(),
                                 covariates=[1.0])
# r.pl = r.hl.plot.manhattan(r.gwas.p_value)
# import bokeh
# bokeh.plotting.show(r.pl)
```
The "QQ plot" that helps evaluate adequacy of the analysis
can be formed using `hl.plot.qq` for very large applications;
here we collect the results for plotting in R.

First we estimate $\lambda_{GC}$
```{r dolam, eval=FALSE}
pv = gwas$p_value$collect()
x2 = stats::qchisq(1-pv,1)
lam = stats::median(x2, na.rm=TRUE)/stats::qchisq(.5,1)
lam
```
And the qqplot:
```{r doqq, eval=FALSE}
qqplot(-log10(ppoints(length(pv))), -log10(pv), xlim=c(0,8), ylim=c(0,8),
  ylab="-log10 p", xlab="expected")
abline(0,1)
```

There is hardly any point to examining a Manhattan plot in
this situation.  But let's see how it might be done.
We'll use igvR to get an interactive and extensible display.

```{r eval=FALSE}
locs <- gwas$locus$collect()
conts <- sapply(locs, function(x) x$contig)
pos <- sapply(locs, function(x) x$position)
library(igvR)
mytab <- data.frame(chr=as.character(conts), pos=pos, pval=pv)
gt <- GWASTrack("simp", mytab, chrom.col=1, pos.col=2, pval.col=3)
igv <- igvR()
setGenome(igv, "hg19")
displayTrack(igv, gt)
```

## Evaluating population stratification

An approach to assessing population stratification
is provided as `hwe_normalized_pca`.  See
the hail [methods docs](https://hail.is/docs/0.2/methods/genetics.html#hail.methods.hwe_normalized_pca)
for details.

We are avoiding a tuple assignment in the tutorial document.
```{python dopca, eval=FALSE}
r.pcastuff = r.hl.hwe_normalized_pca(r.mt.GT)
r.mt = r.mt.annotate_cols(scores=r.pcastuff[1][r.mt.s].scores)
```

We'll collect the key information and plot.
```{r lkpcs, eval=FALSE}
sc <- pcastuff[[2]]$scores$collect()
pc1 = sapply(sc, "[", 1)
pc2 = sapply(sc, "[", 2)
fac = mt$pheno$SuperPopulation$collect()
myd = data.frame(pc1, pc2, pop=fac)
library(ggplot2)
ggplot(myd, aes(x=pc1, y=pc2, colour=factor(pop))) + geom_point()
```

Now repeat the association test with adjustments for
population of origin and gender.

```{python redog, eval=FALSE}
r.gwas2 = r.hl.linear_regression_rows(
    y=r.mt.pheno.CaffeineConsumption,
    x=r.mt.GT.n_alt_alleles(),
    covariates=[1.0,r.mt.pheno.isFemale,r.mt.scores[0],
        r.mt.scores[1], r.mt.scores[2]])
```
New value of $\lambda_{GC}$:

```{r dolam2, eval=FALSE}
pv = gwas2$p_value$collect()
x2 = stats::qchisq(1-pv,1)
lam = stats::median(x2, na.rm=TRUE)/stats::qchisq(.5,1)
lam
```
A manhattan plot for chr8:
```{r doman, eval=FALSE}
locs <- gwas2$locus$collect()
conts <- sapply(locs, function(x) x$contig)
pos <- sapply(locs, function(x) x$position)
mytab <- data.frame(chr=as.character(conts), pos=pos, pval=pv)
ggplot(mytab[mytab$chr=="8",], aes(x=pos, y=-log10(pval))) + geom_point()
```

# Conclusions

The tutorial document proceeds with some illustrations of
arbitrary aggregations.  We will skip these for now.

Additional vignettes will address

- A more realistic higher-volume VCF
- Working with UKBB Summary statistics in GCP
    - https://pan.ukbb.broadinstitute.org/docs/hail-format/index.html
- Representing linkage disequilibrium
    - https://pan-dev.ukbb.broadinstitute.org/docs/ld/index.html
- Simulating variant collections using Balding-Nichols
- Simulating variant collections using Pritchard-Stephens-Donnelly
- Connecting genotypes with phenotype data in FHIR


# SessionInfo {-}

```{r sessionInfo}
sessionInfo()
```

```{r getann}
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute=TRUE)$key_by("Sample")
```
```{python ch3, eval=TRUE}
r.tab.describe()# python chunk!
r.tab.show(width=100)
```

## Adding the sample annotation to the MatrixTable; aggregation

We combine the `tab` defined above, with the MatrixTable instance,
usiny python code reaching to R via `r.`.
```{python ch4, eval=TRUE}
r.mt17 = r.mt17.annotate_cols(pheno = r.tab[r.mt17.s])   # python chunk!
r.mt17.col.describe()
```

Aggregation methods can be used to obtain contingency tables or descriptive statistics.

First, we get the frequencies of superpopulation membership:
```{r getsup, eval=TRUE}
mt17$aggregate_cols(hl$agg$counter(mt17$pheno$SuperPopulation))
```

Then statistics on caffeine consumption:
```{r lkcaf, eval=TRUE}
uu <- mt17$aggregate_cols(hl$agg$stats(mt17$pheno$CaffeineConsumption))
names(uu)
uu$mean
uu$stdev
```

## Evaluating population stratification

An approach to assessing population stratification
is provided as `hwe_normalized_pca`.  See
the hail [methods docs](https://hail.is/docs/0.2/methods/genetics.html#hail.methods.hwe_normalized_pca)
for details.

We are avoiding a tuple assignment in the tutorial document.
```{python dopca, eval=TRUE}
r.pcastuff = r.hl.hwe_normalized_pca(r.mt17.GT)
r.mt17 = r.mt17.annotate_cols(scores=r.pcastuff[1][r.mt17.s].scores)
```

We'll collect the key information and plot.
```{r lkpcs, eval=TRUE}
sc <- pcastuff[[2]]$scores$collect()
pc1 = sapply(sc, "[", 1)
pc2 = sapply(sc, "[", 2)
fac = mt17$pheno$SuperPopulation$collect()
myd = data.frame(pc1, pc2, pop=fac)
library(ggplot2)
ggplot(myd, aes(x=pc1, y=pc2, colour=factor(pop))) + geom_point()
```

# Overview
The purpose of this workspace is to illustrate interoperation of Bioconductor, R, and Spark in AnVIL.  We used
a startup script to install R in a single spark node.  See the Appendix 1 to this description for the startup script, which uses a debian packaging of R 4.2.2, thanks to Dirk Eddelbuettel.

# What to do

## Improve extensibility

The infrastructure is relatively primitive -- we have a version of R 4.2.2 but no associated binary repository.  It should be possible to use the r2u package management system to get CRAN packages without compilation but this has not been attempted as of 20 Jan 2023.

## Use BiocHail

### Basic setup

We can use `install.packages(c("BiocManager", "remotes"))` and then `BiocManager::install("vjcitn/BiocHail")` to get everything needed to use Hail with Bioconductor under the basilisk discipline, with all python dependencies specified.

**Crucial step before using `hail_init`: ensure that environment variable PIP_USER is set to false** and that this is known to the R session in which `hail_init` is first called.  If this is done, all dependencies will be consistently installed under `$HOME/.cache/R/basilisk` at the time `hail_init()` is run.  This step generates a persistent installation that will not need to be revised unless a new version of basilisk or BiocHail are installed in R.

### Work with the hail.is extract of 1000 genomes data in GRCh37 reference

We don't see a way to add an R kernel to jupyter in this cloud environment, although perhaps the startup script could be emended to deal with that.  We'll have to work in the terminal.

Here's how we acquire and inspect the 1000 genomes subset from hail.is for the GWAS tutorial.  This creates
a persistent cache where the `get_1kg()` function can find the resource when requested in future sessions.

```
> library(BiocHail)
> hl = hail_init()
> my = get_1kg(hl)
/home/jupyter/.cache/R/BiocFileCache
  does not exist, create directory? (yes/no): yes
  |======================================================================| 100%

> my
<hail.matrixtable.MatrixTable object at 0x7f76c2e7e4f0>
> my$rows()$select()$show(5L)
+---------------+------------+                                                                                              (0 + 2) / 2]
| locus         | alleles    |
+---------------+------------+
| locus<GRCh37> | array<str> |
+---------------+------------+
| 1:904165      | ["G","A"]  |
| 1:909917      | ["G","A"]  |
| 1:986963      | ["C","T"]  |
| 1:1563691     | ["T","G"]  |
| 1:1707740     | ["T","G"]  |
+---------------+------------+
showing top 5 rows
```
We can annotate the `my` MatrixTable instance, getting a new object `mt`.  First
acquire the annotation table for the tutorial:
```
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute=TRUE)$key_by("Sample")
```
Then add to the genotype table:

```
mt = my$annotate_cols(pheno = tab[my$s])
```
This works nicely --
```
> mt$col$describe()
--------------------------------------------------------
Type:
        struct {
        s: str,
        pheno: struct {
            Population: str,
            SuperPopulation: str,
            isFemale: bool,
            PurpleHair: bool,
            CaffeineConsumption: int32
        }
    }
--------------------------------------------------------
Source:
    <hail.matrixtable.MatrixTable object at 0x7f76c2493250>
Index:
    ['column']
--------------------------------------------------------
> mt$aggregate_cols(hl$agg$counter(mt$pheno$SuperPopulation))
$AFR
[1] 76

$AMR
[1] 34

$EAS
[1] 72

$EUR
[1] 47

$SAS
[1] 55
```

At this point it should be clear how to carry out all the steps of the [tutorial in R](https://vjcitn.github.io/BiocHail/01_gwas_tut.html).  It will be much nicer when we have an R kernel in jupyter with Spark/Hail.

### Work with 1000 genomes data called with the Telomere-to-telomere reference


We used
```
wget https://mghp.osn.xsede.org/bir190004-bucket01/Bioc1KGt2t/t17.zip
```
to acquire a 45GB zip file from Bioconductor's Open Storage Network allocation.  This zip
file was generated from the VCF file for chr17 of the T2T genotype calls; see Appendix 2 for details.

```
> Sys.setenv(HAIL_T2T_CHR17="/home/jupyter/t2t17.mt")
> hl <- hail_init()
> # NB the following two commands are now encapsulated in the rg_update function
> nn <- hl$get_reference('GRCh38')
> nn <- nn$read(system.file("json/t2tAnVIL.json", package="BiocHail"))
> # updates the hail reference genome
> bigloc = Sys.getenv("HAIL_T2T_CHR17")
> if (nchar(bigloc)>0) {
+   mt17 <- hl$read_matrix_table(Sys.getenv("HAIL_T2T_CHR17"))
+   mt17$count()
+ }
[[1]]
[1] 3824434

[[2]]
[1] 3202
```

At this point one can follow the code in the [vignette on working with chr17 for the 3202
1KG samples in T2T reference.](https://vjcitn.github.io/BiocHail/02_large_t2t.html).

# Appendix 1 -- establishing R in a terra dataproc cluster
```
!/usr/bin/env bash

# 12/27/2022 will produce R 4.2.2 in docker ubuntu:22.04 in 2 minutes

# added
apt update
apt install --yes ca-certificates
DEBIAN_FRONTEND=noninteractive apt-get install -y tzdata

apt install --yes --no-install-recommends wget          # to add the key
wget -q -O- https://eddelbuettel.github.io/r2u/assets/dirk_eddelbuettel_key.asc \
    | tee -a /etc/apt/trusted.gpg.d/cranapt_key.asc

# USE FOCAL

#echo "deb [arch=amd64] https://dirk.eddelbuettel.com/cranapt focal main" \
#    > /etc/apt/sources.list.d/cranapt.list
#apt update
# or use the mirror at the University of Illinois Urbana-Champaign:

echo "deb [arch=amd64] https://r2u.stat.illinois.edu/ubuntu focal main" \
    > /etc/apt/sources.list.d/cranapt.list
apt update

# (In either example, replace focal with focal for use with Ubuntu 22.04.)

#Third, and optionally, if you do not yet have the current R version, run these two lines (or use the standard CRAN repo setup)
#

wget -q -O- https://cloud.r-project.org/bin/linux/ubuntu/marutter_pubkey.asc \
    | tee -a /etc/apt/trusted.gpg.d/cran_ubuntu_key.asc
echo "deb [arch=amd64] https://cloud.r-project.org/bin/linux/ubuntu focal-cran40/" \
    > /etc/apt/sources.list.d/cran-ubuntu.list
apt update
apt install -y r-base-core
```

# Appendix 2 - producing a MatrixTable with 1000 Genomes data genotyped to the T2T reference
We localized the raw genotypes to an AnVIL workspace disk:
```
AnVIL::gsutil_cp("gs://fc-47de7dae-e8e6-429c-b760-b4ba49136eee/1KGP/joint_genotyping/joint_vcfs/raw/chr17.genotyped.vcf.gz", ".")
```

We created a JSON file with the chromosome lengths of the T2T reference in t2tAnVIL.json.
We then used python in a large Hail/Spark cluster in AnVIL:
```
>>> import hail as h
>>> rg = h.get_reference('GRCh38')
Initializing Hail with default parameters...
Setting default log level to "WARN".
To adjust logging level use sc.setLogLevel(newLevel). For SparkR, use setLogLevel(newLevel).
Running on Apache Spark version 3.1.3
SparkUI available at http://756809c79837:4040
Welcome to
     __  __     <>__
    / /_/ /__  __/ /
   / __  / _ `/ / /
  /_/ /_/\_,_/_/_/   version 0.2.105-acd89e80c345
LOGGING: writing to /home/rstudio/hail-20221213-1558-0.2.105-acd89e80c345.log
>>> nn = rg.read("t2tAnVIL.json")
>>> h.import_vcf("chr17.genotyped.vcf.gz", force_bgz=True, reference_genome=nn).write("t2t17.mt")
```
This operation seems to take a long time even with 64 cores.  We
could not get exact timing owing to some connectivity problems.

This t2t17.mt was then zipped up and copied to Bioconductor's Open Storage Network allocation.  In principle this file could be placed in Zenodo for long term access.

---
title: "02 Working with larger VCF: T2T by chromosome"
author: "Vincent J. Carey, stvjc at channing.harvard.edu"
date: "`r format(Sys.time(), '%B %d, %Y')`"
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{02 Working with larger VCF: T2T by chromosome}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document:
    highlight: pygments
    number_sections: yes
    theme: united
    toc: yes
---

# Overview

In this document we illustrate some issues with
large data volumes.

Here is how we acquire the genotypes for the thousand
genomes samples based on the T2T reference.  We obtained
bgzipped vcf via
```
AnVIL::gsutil_cp("gs://fc-47de7dae-e8e6-429c-b760-b4ba49136eee/1KGP/joint_genotyping/joint_vcfs/raw/chr22.genotyped.vcf.gz", ".")
```
Then we used hail in python to deal with the conversion to MatrixTable.
We could have done this in R, but we had to learn how to manipulate
the 'reference sequence'.  We have a conjectural 'reference sequence'
json document in the json folder of the BiocHail package, used here
as `t2tAnVIL.json`.
```
>>> import hail as h
>>> rg = h.get_reference('GRCh38')
Initializing Hail with default parameters...
Setting default log level to "WARN".
To adjust logging level use sc.setLogLevel(newLevel). For SparkR, use setLogLevel(newLevel).
Running on Apache Spark version 3.1.3
SparkUI available at http://756809c79837:4040
Welcome to
     __  __     <>__
    / /_/ /__  __/ /
   / __  / _ `/ / /
  /_/ /_/\_,_/_/_/   version 0.2.105-acd89e80c345
LOGGING: writing to /home/rstudio/hail-20221213-1558-0.2.105-acd89e80c345.log
>>> nn = rg.read("t2tAnVIL.json")
>>> h.import_vcf("chr22.genotyped.vcf.gz", force_bgz=True, reference_genome=nn).write("t2t22.mt")
```
This operation seems to take a long time even with 64 cores.  We
could not get exact timing owing to some connectivity problems.

Here we'll work with the genotype data for T2T chr17.  We assume
that the MatrixTable is located at the path given by the environment
variable `HAIL_T2T_CHR17`.  This MatrixTable is available
in the Open Storage Network at osn:/bir190004-bucket01/Bioc1KGt2t/t17.zip.
This is a 45 GB file.  It should be unzipped at the location pointed to
by `HAIL_T2T_CHR17`.  Instructions for using rclone to acquire the zip file
are given in the appendix.

```{r doini,message=FALSE}
library(BiocHail)
```

(If the above command indicates that BiocHail is not available,
see the Installation section near the end of this document.)

```{r do17, eval=TRUE}
hl <- hail_init()
# NB the following two commands are now encapsulated in the rg_update function
nn <- hl$get_reference('GRCh38')
nn <- nn$read(system.file("json/t2tAnVIL.json", package="BiocHail"))
# updates the hail reference genome
bigloc = Sys.getenv("HAIL_T2T_CHR17")
if (nchar(bigloc)>0) {
  mt17 <- hl$read_matrix_table(Sys.getenv("HAIL_T2T_CHR17"))
  mt17$count()
}
```

# Population stratification assessment via PCA

The following command would compute PCs based on all SNP.
```
pcastuff = hl$hwe_normalized_pca(mt17$GT)
```
This seems impractical.  We have found that with a 64 core machine
at terra.bio, PCA on samples of 1-5% of the 3.8 million loci on T2T chr17
takes 40 minutes.  Hail's code seems good at utilizing all the cores.

We saved the PC scores for PCA based on 38k randomly sampled loci,
and 191k randomly sampled loci.  Here's a simple view of the latter.

```{r lk191,eval=TRUE}
utils::data(pcs_191k)
graphics::pairs(pcs_191k[,1:5], pch=".")
```

# Exercises

1. Comment on the gain in information about geographic origin
achieved by using a 5% sample of loci instead of a 1% sample.

2. Find the geographic origins of donors of the 1000 genomes genotypes
and bind them to `mt17` using the methods given in vignette `01_gwas_tut`.
Use these codes to color the points in the PCA plot.

3. Produce an artificial "phenotype" for these donors via `rnorm(3202,0,1)`,
bind it to the genotype data, and perform a naive GWAS.  What are the
loci most strongly associated with the artificial phenotype?

4. Produce a new artificial phenotype which has some association with
geographic origin of donor, but no association with genotype.
Produce a new naive GWAS, and a third using some of the PCA scores
as covariates.  What are the effects of this covariate adjustment on
reasoning about genetic association with the artificial phenotype?


# Appendix: using rclone with docker to get the chr17 data

It can be painful to install and configure rclone.  We use
a docker container.  Let RC_DATADIR be an environment variable
evaluating to an available folder.  

Also, place the text file with contents
```
[osn]
type = s3
provider = AWS
endpoint = https://mghp.osn.xsede.org
acl = public
no_check_bucket = true
```
in a file `rclone.conf` in a folder pointed to by the environment
variable `RC_CONFDIR`.


Then the following

```
docker run -v $RC_DATADIR:/data -v $RC_CONFDIR:/config/rclone -t rclone/rclone:latest ls osn:/bir190004-bucket01/Bioc1KGt2t
```

will list the files with 1KG samples genotyped against the T2T reference.



Use the rclone `copyto` command to obtain a local copy of the zip file `t17.zip`
in the folder pointed to by `$RC_DATADIR`:

```
docker run -v $RC_DATADIR:/data -v $RC_CONFDIR:/config/rclone -t rclone/rclone:latest copyto osn:/bir190004-bucket01/Bioc1KGt2t/t17.zip ./t17.zip
```
This file should be unzipped in a folder to which the environment variable `HAIL_T2T_CHR17` will
point.

# Installing `BiocHail`

`BiocHail` should be installed as follows:

```{r, eval = FALSE}
if (!require("BiocManager"))
    install.packages("BiocManager")
BiocManager::install("BiocHail")
```

As of 1.0.0, a JDK for Java version `<=` 11.0 is necessary
to use the version of Hail that is installed with the package.
This package should be usable on MacOS with suitable java
support.  If Java version `>=` 8.x is used, warnings from
Apache Spark may be observed.  To the best of our knowledge
the conditions to which the warnings pertain do not affect program performance.


# SessionInfo {-}

```{r sessionInfo}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{pcs_38k}
\alias{pcs_38k}
\title{HWE-normalized PCA scores for 3202 thousand-genomes samples genotyped with the
telomere-to-telomere reference}
\format{
data.frame
}
\usage{
data("pcs_38k")
}
\value{
data.frame
}
\description{
HWE-normalized PCA scores for 3202 thousand-genomes samples genotyped with the
telomere-to-telomere reference
}
\note{
The genotypes are from a 1% sample of loci on chr17
}
\examples{
data(pcs_38k)
dim(pcs_38k)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rg_update.R
\name{rg_update}
\alias{rg_update}
\title{update the reference genome for a hail instance}
\usage{
rg_update(
  hc,
  init = "GRCh38",
  newjson = system.file("json/t2tAnVIL.json", package = "BiocHail")
)
}
\arguments{
\item{hc}{hail context}

\item{init}{character(1) valid name for a reference genome, defaults to "GRCh38"}

\item{newjson}{character(1) path to a json spec of a reference genome [needs doc]}
}
\value{
a python list; the function is used for its side effect
}
\description{
update the reference genome for a hail instance
}
\examples{
hl <- hail_init()
rg_update(hl)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ukbb_init.R
\name{ukbb_init}
\alias{ukbb_init}
\title{initialize ukbb}
\usage{
ukbb_init()
}
\value{
python module reference
}
\description{
initialize ukbb
}
\note{
ukbb module may be passed around
}
\examples{
ukbb <- ukbb_init()
names(ukbb)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.data.frame.R
\name{as.data.frame.hail.table.Table}
\alias{as.data.frame.hail.table.Table}
\title{convert hail.table.Table to R data frame}
\usage{
\method{as.data.frame}{hail.table.Table}(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{x}{instance of "hail.table.Table"}

\item{row.names}{not used}

\item{optional}{not used}

\item{\dots}{not used}
}
\value{
data.frame
}
\description{
convert hail.table.Table to R data frame
}
\note{
only use on small table because collect is used
}
\examples{
hl <- hail_init()
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
as.data.frame(tab$head(3L))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hail_control.R
\name{hail_init_simple}
\alias{hail_init_simple}
\title{initialize hail}
\usage{
hail_init_simple()
}
\value{
python reference to hail module
}
\description{
initialize hail
}
\note{
hail object may be passed around
}
\examples{
hc <- hail_init_simple()
hc
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rownames.R
\name{rownames,hail.table.Table-method}
\alias{rownames,hail.table.Table-method}
\title{acquire row names of a Hail Table, assuming key has been set}
\usage{
\S4method{rownames}{hail.table.Table}(x, do.NULL = TRUE, prefix = "row")
}
\arguments{
\item{x}{instance of hail.table.Table}

\item{do.NULL}{not used}

\item{prefix}{not used}
}
\value{
character()

character vector
}
\description{
acquire row names of a Hail Table, assuming key has been set
}
\note{
To try example, run `example("rownames,hail.table.Table-method")`
}
\examples{
hl <- hail_init()
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
rt <- rownames(tab)
length(rt)
head(rt)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_1kg.R
\name{osn_1kg_path}
\alias{osn_1kg_path}
\title{Open Storage Network path to a zip of hail MatrixTable with some 1kg data for the Hail.is GWAS tutorial}
\usage{
osn_1kg_path()
}
\value{
character(1) URL to zip
}
\description{
Open Storage Network path to a zip of hail MatrixTable with some 1kg data for the Hail.is GWAS tutorial
}
\examples{
osn_1kg_path()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pheno_col_ops.R
\name{multipop_df}
\alias{multipop_df}
\title{pheno_data component harvesting from columns of summary stats MatrixTable
allowing for info on multiple populations in the pheno_data component}
\usage{
multipop_df(
  x,
  top2get = c("trait_type", "phenocode", "description", "modifier", "coding_description",
    "coding"),
  pheno2get = c("n_cases", "n_controls", "heritability", "pop")
)
}
\arguments{
\item{x}{Struct - a single element of the list returned by mt$cols()$collect()}

\item{top2get}{character() vector of general fields to retrieve}

\item{pheno2get}{character() vector of fields to be retrieved for each subpopulation}
}
\value{
data.frame
}
\description{
pheno_data component harvesting from columns of summary stats MatrixTable
allowing for info on multiple populations in the pheno_data component
}
\examples{
# following are too time-consuming but can be of interest
# if (nchar(Sys.getenv("HAIL_UKBB_SUMSTAT_10K_PATH"))>0) {
#  hl = hail_init()
#  ss = get_ukbb_sumstat_10kloci_mt(hl)
#  sscol = ss$cols()$collect() # may take a bit of time
#  print(length(sscol))
#  multipop_df(sscol[[1]])
# }
#
\donttest{
# if (nchar(Sys.getenv("HAIL_UKBB_SUMSTAT_10K_PATH"))>0) {
# # to get an overview of all phenotype-cohort combinations in a searchable table
# mmm = lapply(sscol, multipop_df )
# mymy = do.call(rbind, mmm) # over 16k rows
# DT::datatable(mymy)
# }
#
}
# this runs quickly and is demonstrative
hl <- hail_init()
litzip <- system.file("extdata", "myss2.zip", package = "BiocHail")
td <- tempdir()
unzip(litzip, exdir = td)
ntab <- hl$read_matrix_table(paste0(td, "/myss2.mt"))
ntab$describe()
nt2 <- ntab$col$collect()
multipop_df(nt2[[1]]) # must select one element
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_1kg.R
\name{path_1kg_annotations}
\alias{path_1kg_annotations}
\title{generate path to installed annotations file}
\usage{
path_1kg_annotations()
}
\value{
character(1) path to annotations
}
\description{
generate path to installed annotations file
}
\note{
.txt file retrieved from extraction on `https://storage.googleapis.com/hail-1kg/tutorial_data.tar`
}
\examples{
path_1kg_annotations()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{kg_3202}
\alias{kg_3202}
\title{data.frame with metadata about 3202 samples genotyped against T2T reference}
\format{
data.frame
}
\usage{
data("kg_3202")
}
\value{
data.frame
}
\description{
data.frame with metadata about 3202 samples genotyped against T2T reference
}
\note{
Source: index files described at `https://www.internationalgenome.org/data-portal/data-collection/30x-grch38`
}
\examples{
data(kg_3202)
dim(kg_3202)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hail_control.R
\name{bare_hail}
\alias{bare_hail}
\title{bare interface to hail using reticulate}
\usage{
bare_hail()
}
\value{
python reference to hail module
}
\description{
bare interface to hail using reticulate
}
\note{
`/home/jupyter/.local/share/r-miniconda/envs/r-reticulate/bin/pip3 install...` is 
used to ensure that reticulate's python ecosystem is what we want
}
\examples{
# assumes terra
if (nchar(Sys.getenv("WORKSPACE_NAMESPACE"))>0) {
  hl = bare_hail()
  hl$init(idempotent=TRUE, spark_conf=list(
      'spark.hadoop.fs.gs.requester.pays.mode'= 'CUSTOM',
      'spark.hadoop.fs.gs.requester.pays.buckets'= 'ukb-diverse-pops-public',
      'spark.hadoop.fs.gs.requester.pays.project.id'= Sys.getenv("GOOGLE_PROJECT")))
  hl$read_matrix_table('gs://ukb-diverse-pops-public/sumstats_release/results_full.mt')$describe()
  \dontrun{
  # this is supposed to get us some LD data but xx.shape fails, issue filed
    hli = reticulate::import("hail.linalg")
    upa = reticulate::import("ukbb_pan_ancestry")
    xx = hli$BlockMatrix$read(upa$get_ld_matrix_path('AFR'))
   }
  }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hail_control.R
\name{hail_init}
\alias{hail_init}
\title{initialize hail, using more options}
\usage{
hail_init(
  quiet = FALSE,
  min_block_size = 0L,
  branching_factor = 50L,
  default_reference = "GRCh37",
  global_seed = 1234L,
  log = tempfile(),
  spark_conf = NULL,
  gcs_requester_pays_configuration = NULL
)
}
\arguments{
\item{quiet}{logical(1) defaults to FALSE}

\item{min_block_size}{integer(1) defaults to 0L}

\item{branching_factor}{integer(1) defaults to 50L}

\item{default_reference}{character(1) defaults to "GRCh37", for compatibility with earlier `hail_init`}

\item{global_seed}{integer(1) defaults to 1234L}

\item{log}{character(1) target folder for logging, defaults to tempfile()}

\item{spark_conf}{list, defaults to NULL}

\item{gcs_requester_pays_configuration}{list, defaults to NULL}
}
\value{
python reference to hail module
}
\description{
initialize hail, using more options
}
\note{
hail object may be passed around.  See hail documentation for details on all args.
}
\examples{
proj = Sys.getenv("GOOGLE_PROJECT")
buck = Sys.getenv("GCS_BUCKET")
if (nchar(buck)>0) {
  # conf = list(proj, c(buck)) doesn't seem to generate tuple[str,Sequence[str]]
  hl <- hail_init()   #gcs_requester_pays_configuration=conf)
  hl$default_reference()
}  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ukbb_sumst.R
\name{osn_ukbb_sumst10k_path}
\alias{osn_ukbb_sumst10k_path}
\title{Open Storage Network path to a zip of hail MatrixTable with a small subset of UKBB summary statistics as of 12/25/2022}
\usage{
osn_ukbb_sumst10k_path()
}
\value{
character(1) path to zip
}
\description{
Open Storage Network path to a zip of hail MatrixTable with a small subset of UKBB summary statistics as of 12/25/2022
}
\examples{
osn_ukbb_sumst10k_path()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pheno_col_ops.R
\name{top2df}
\alias{top2df}
\title{top-level annotation harvesting from columns of summary statistics MatrixTable}
\usage{
top2df(
  x,
  toget = c("trait_type", "phenocode", "description", "modifier", "coding_description")
)
}
\arguments{
\item{x}{a Struct returned from mt$cols()$collect() -- which can be slow}

\item{toget}{character() vector of field names to retrieve}
}
\value{
1-row data.frame
}
\description{
top-level annotation harvesting from columns of summary statistics MatrixTable
}
\note{
python None are transformed to NA
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_1kg.R
\name{get_1kg}
\alias{get_1kg}
\title{interface to 1kg import}
\usage{
get_1kg(
  hl,
  retrieve_import_write = FALSE,
  path_1kg_zip = osn_1kg_path(),
  folder = tempdir(),
  cache = BiocFileCache::BiocFileCache()
)
}
\arguments{
\item{hl}{hail object}

\item{retrieve_import_write}{logical(1) if TRUE, use hl.utils.get_1kg to retrieve data, otherwise acquire
a previously written zip file, either from a cache, or, if no file found in cache, from web, followed by caching}

\item{path_1kg_zip}{character(1) path to zip of MatrixTable, defaults to `osn_1kg_path()`.}

\item{folder}{character(1) destination of 1kg.mt as retrieved using hl.utils.get_1kg, import_vcf, write}

\item{cache}{a BiocFileCache-type cache}
}
\value{
"hail.matrixtable.MatrixTable" instance
}
\description{
interface to 1kg import
}
\note{
overwrite is permitted in the import_vcf.write event
}
\examples{
hl <- hail_init()
mt <- get_1kg(hl)
mt
mt$rows()$select()$show(5L) # must use integer
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
tab$describe()
tab$show(width = 100L)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pheno_col_ops.R
\name{pheno_data_sec_2df}
\alias{pheno_data_sec_2df}
\title{pheno_data component harvesting from columns of summary stats MatrixTable}
\usage{
pheno_data_sec_2df(
  m,
  section = 1,
  toget = c("n_cases", "n_controls", "heritability", "pop"),
  verbose = FALSE
)
}
\arguments{
\item{m}{Struct returned from mt$cols()$collect()}

\item{section}{numeric(1) element of pheno_data list to be transformed to data.frame}

\item{toget}{character() vector of field names to retrieve}

\item{verbose}{logical(1) if TRUE (NOT default) will message that there are multiple
`pheno_data` components returned}
}
\value{
1 row data.frame
}
\description{
pheno_data component harvesting from columns of summary stats MatrixTable
}
\note{
applies top2df to the pheno_data component of input
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colnames.R
\name{get_key.hail.table.Table}
\alias{get_key.hail.table.Table}
\title{S3 method for get_key}
\usage{
\method{get_key}{hail.table.Table}(x)
}
\arguments{
\item{x}{instance of hail.table.Table}
}
\value{
a list with elements names (names of keys) and key_df (data.frame of key values, with column names)
}
\description{
S3 method for get_key
}
\examples{
hl <- hail_init()
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
get_key(tab)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colnames.R
\name{get_key}
\alias{get_key}
\title{S3 generic for get_key}
\usage{
get_key(x)
}
\arguments{
\item{x}{anything}
}
\value{
typically a list
}
\description{
S3 generic for get_key
}
\examples{
hl <- hail_init()
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
get_key(tab)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/table_filter.R
\name{filter}
\alias{filter}
\title{s3 support}
\usage{
filter(.data, ..., .by = NULL, .preserve = FALSE)
}
\arguments{
\item{.data}{instance of hail.table.Table}

\item{\dots}{should include named components `filter` which is a logical vector
with same number of rows as `.data`, `hl`, a reference to a hail environment (Module),
and `placeholder` an arbitrary character(1)}

\item{.by}{not used}

\item{.preserve}{not used}
}
\value{
filtered hail.table.Table reference
}
\description{
s3 support
}
\examples{
hl <- hail_init()
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
pick <- rep(FALSE, 3500)
pick[seq_len(10)] <- TRUE
ft <- filter(tab, filter = pick, hl = hl)
ft$count()
ft$head(2L)$collect()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.data.frame.R
\name{as.data.frame}
\alias{as.data.frame}
\title{S3 support}
\usage{
as.data.frame(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{x}{entity coercible to data.frame}

\item{row.names}{character or NULL}

\item{optional}{logical}

\item{\dots}{any args}
}
\value{
data.frame
}
\description{
S3 support
}
\examples{
hl <- hail_init()
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
as.data.frame(tab$head(3L))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_ukbb_sumst.R
\name{get_ukbb_sumstat_10kloci_mt}
\alias{get_ukbb_sumstat_10kloci_mt}
\title{interface to a small subset of UKBB summary stats in MatrixTable format}
\usage{
get_ukbb_sumstat_10kloci_mt(
  hl,
  folder = tempdir(),
  cache = BiocFileCache::BiocFileCache(),
  timeout.ukbb = 3600
)
}
\arguments{
\item{hl}{hail object}

\item{folder}{character(1) destination of 1kg.mt as retrieved using hl.utils.get_1kg, import_vcf, write}

\item{cache}{a BiocFileCache-type cache}

\item{timeout.ukbb}{numeric(1) defaults to 3600 for timeout setting in `options()`; option value is reset on exit}
}
\value{
"hail.matrixtable.MatrixTable" instance
}
\description{
interface to a small subset of UKBB summary stats in MatrixTable format
}
\note{
The loci were selected using a .000345% random sample
of over 28 million loci recorded in the UKBB pan-ancestry
record.  The sample is made available a) to assess some issues with data volume (the full
resource is about 12.78 TB according to [this doc](https://pan-dev.ukbb.broadinstitute.org/docs/hail-format/index.html), and b) to provide full information on the scope of phenotypes and populations available.

This function will unzip 5GB of MatrixTable data.  It may be desirable to cache the
unzipped image to a persistent location.  If this has been done and the environment variable
`HAIL_UKBB_SUMSTAT_10K_PATH` has been set to this location, this function will use
the MatrixTable content found there.
}
\examples{
hl <- hail_init()
ss <- get_ukbb_sumstat_10kloci_mt(hl)
# consider saving the unzipped image and recaching
ss$count()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colnames.R
\name{colnames,hail.table.Table-method}
\alias{colnames,hail.table.Table-method}
\title{extract field names from hail.table.Table}
\usage{
\S4method{colnames}{hail.table.Table}(x, do.NULL = TRUE, prefix = "col")
}
\arguments{
\item{x}{hail.table.Table instance}

\item{do.NULL}{ignored}

\item{prefix}{ignored}
}
\value{
character vector
}
\description{
extract field names from hail.table.Table
}
\examples{
hl <- hail_init()
annopath <- path_1kg_annotations()
tab <- hl$import_table(annopath, impute = TRUE)$key_by("Sample")
colnames(tab)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hail_control.R
\name{hail_stop}
\alias{hail_stop}
\title{stop hail}
\usage{
hail_stop(hl)
}
\arguments{
\item{hl}{a hail object produced by hail_init()}
}
\value{
result of stop() method for Hail module
}
\description{
stop hail
}
\examples{
hail_stop
}
